@@ -5,11 +5,9 @@ from collections import defaultdict
 from contextlib import contextmanager
 import itertools
 import re
-from typing import (Any, Callable, Dict, Iterable, Iterator, List, Optional,
-                    Set, TYPE_CHECKING, Tuple, TypeVar)
+from typing import Any, Callable, Iterable, Iterator, TYPE_CHECKING, TypeVar
 
-
-from langkit.diagnostics import check_source_language, error
+from langkit.diagnostics import Location, check_source_language, error
 from langkit.lexer.char_set import CharSet
 
 
@@ -17,7 +15,7 @@ if TYPE_CHECKING:
     from langkit.lexer import RuleAssoc
 
 
-rule_name_re = re.compile('^[a-zA-Z][a-zA-Z0-9_]*$')
+rule_name_re = re.compile("^[a-zA-Z][a-zA-Z0-9_]*$")
 repeat_re = re.compile("^(?P<low>[0-9]+)(?P<comma>,(?P<high>[0-9]+)?)$")
 
 
@@ -25,9 +23,11 @@ T = TypeVar("T")
 U = TypeVar("U")
 
 
-def _to_dot(starting_state: T,
-            get_transitions: Callable[[T], List[Tuple[U, T]]],
-            get_state_label: Callable[[T], Optional[str]]) -> str:
+def _to_dot(
+    starting_state: T,
+    get_transitions: Callable[[T], list[tuple[U, T]]],
+    get_state_label: Callable[[T], str | None],
+) -> str:
     """
     Helper to emit a dot(1) file representing a states graph.
 
@@ -43,10 +43,10 @@ def _to_dot(starting_state: T,
         state.
     """
     id_generator = iter(itertools.count(0))
-    ids: Dict[T, int] = {}
+    ids: dict[T, int] = {}
 
-    nodes: List[str] = []
-    edges: List[str] = []
+    nodes: list[str] = []
+    edges: list[str] = []
 
     def add_node(n: T) -> int:
         try:
@@ -59,17 +59,18 @@ def _to_dot(starting_state: T,
 
         for label, next_state in get_transitions(n):
             next_id = add_node(next_state)
-            edges.append('{} -> {} [label="{}"];'.format(
-                id, next_id, label))
+            edges.append('{} -> {} [label="{}"];'.format(id, next_id, label))
 
         node_label = get_state_label(n)
-        nodes.append('{} [label="{}"];'.format(
-            id, '{} {}'.format(id, node_label) if node_label else id
-        ))
+        nodes.append(
+            '{} [label="{}"];'.format(
+                id, "{} {}".format(id, node_label) if node_label else id
+            )
+        )
         return id
 
     add_node(starting_state)
-    return '\n'.join(['digraph g {'] + nodes + edges + ['}'])
+    return "\n".join(["digraph g {"] + nodes + edges + ["}"])
 
 
 class SequenceReader:
@@ -105,8 +106,9 @@ class RegexpCollection:
         """Base class for regexp components."""
 
         @abc.abstractmethod
-        def to_nfa(self,
-                   regexps: RegexpCollection) -> Tuple[NFAState, NFAState]:
+        def to_nfa(
+            self, regexps: RegexpCollection
+        ) -> tuple[NFAState, NFAState]:
             """
             Turn this parser into a NFA.
 
@@ -118,11 +120,12 @@ class RegexpCollection:
     class Sequence(Parser):
         """Consume input that sub-parsers can consume sequentially."""
 
-        def __init__(self, subparsers: List[RegexpCollection.Parser]):
+        def __init__(self, subparsers: list[RegexpCollection.Parser]):
             self.subparsers = subparsers
 
-        def to_nfa(self,
-                   regexps: RegexpCollection) -> Tuple[NFAState, NFAState]:
+        def to_nfa(
+            self, regexps: RegexpCollection
+        ) -> tuple[NFAState, NFAState]:
             # If this sequences matches only the empty input, just return a
             # single state.
             if not self.subparsers:
@@ -136,8 +139,9 @@ class RegexpCollection:
             return (sub_nfas[0][0], sub_nfas[-1][1])
 
         def __repr__(self) -> str:
-            return 'Seq({})'.format(', '.join(
-                repr(s) for s in self.subparsers))
+            return "Seq({})".format(
+                ", ".join(repr(s) for s in self.subparsers)
+            )
 
     class Repeat(Parser):
         """Accept input running the sub-parser multiple times."""
@@ -145,8 +149,9 @@ class RegexpCollection:
         def __init__(self, subparser: RegexpCollection.Parser):
             self.subparser = subparser
 
-        def to_nfa(self,
-                   regexps: RegexpCollection) -> Tuple[NFAState, NFAState]:
+        def to_nfa(
+            self, regexps: RegexpCollection
+        ) -> tuple[NFAState, NFAState]:
             # Get the subparser NFA and connect its ends to match repetitions
             nfa = self.subparser.to_nfa(regexps)
             nfa[1].add_transition(None, nfa[0])
@@ -156,16 +161,17 @@ class RegexpCollection:
             return (nfa[0], nfa[0])
 
         def __repr__(self) -> str:
-            return 'Repeat({})'.format(self.subparser)
+            return "Repeat({})".format(self.subparser)
 
     class Or(Parser):
         """Accept input that at least one sub-parser accepts."""
 
-        def __init__(self, subparsers: List[RegexpCollection.Parser]):
+        def __init__(self, subparsers: list[RegexpCollection.Parser]):
             self.subparsers = subparsers
 
-        def to_nfa(self,
-                   regexps: RegexpCollection) -> Tuple[NFAState, NFAState]:
+        def to_nfa(
+            self, regexps: RegexpCollection
+        ) -> tuple[NFAState, NFAState]:
             starting = NFAState()
             ending = NFAState()
 
@@ -180,8 +186,7 @@ class RegexpCollection:
             return (starting, ending)
 
         def __repr__(self) -> str:
-            return 'Or({})'.format(', '.join(
-                repr(s) for s in self.subparsers))
+            return "Or({})".format(", ".join(repr(s) for s in self.subparsers))
 
     class Opt(Parser):
         """Accept input that the sub-parser accepts, or the empty input."""
@@ -189,14 +194,15 @@ class RegexpCollection:
         def __init__(self, subparser: RegexpCollection.Parser):
             self.subparser = subparser
 
-        def to_nfa(self,
-                   regexps: RegexpCollection) -> Tuple[NFAState, NFAState]:
+        def to_nfa(
+            self, regexps: RegexpCollection
+        ) -> tuple[NFAState, NFAState]:
             starting, ending = self.subparser.to_nfa(regexps)
             starting.add_transition(None, ending)
             return (starting, ending)
 
         def __repr__(self) -> str:
-            return 'Opt({})'.format(self.subparser)
+            return "Opt({})".format(self.subparser)
 
     class Range(Parser):
         """Accept any character in the given set."""
@@ -204,8 +210,9 @@ class RegexpCollection:
         def __init__(self, char_set: CharSet):
             self.char_set = char_set
 
-        def to_nfa(self,
-                   regexps: RegexpCollection) -> Tuple[NFAState, NFAState]:
+        def to_nfa(
+            self, regexps: RegexpCollection
+        ) -> tuple[NFAState, NFAState]:
             starting = NFAState()
             ending = NFAState()
             starting.add_transition(self.char_set, ending)
@@ -220,28 +227,49 @@ class RegexpCollection:
         def __init__(self, name: str):
             self.name = name
 
-        def to_nfa(self,
-                   regexps: RegexpCollection) -> Tuple[NFAState, NFAState]:
+        def to_nfa(
+            self, regexps: RegexpCollection
+        ) -> tuple[NFAState, NFAState]:
             with regexps._visit_rule(self.name):
                 try:
                     parser = regexps.patterns[self.name]
                 except KeyError:
-                    check_source_language(
-                        False, 'unknown pattern: {}'.format(self.name))
+                    error(
+                        f"unknown pattern: {self.name}",
+                        location=RegexpCollection._current_location,
+                    )
                 return parser.to_nfa(regexps)
 
         def __repr__(self) -> str:
-            return '@{}'.format(self.name)
+            return "@{}".format(self.name)
 
     escape_chars = {
-        '(': '(', ')': ')',
-        '[': '[', ']': ']',
-        '|': '|', '+': '+', '*': '*', '\\': '\\',
-        'a': '\a', 'b': '\b', 'B': '\\', 'e': '\033', 'f': '\f',
-        'n': '\n', 'r': '\r', 't': '\t', 'v': '\v',
-        '0': '\0',
+        "(": "(",
+        ")": ")",
+        "[": "[",
+        "]": "]",
+        "|": "|",
+        "+": "+",
+        "*": "*",
+        "\\": "\\",
+        "a": "\a",
+        "b": "\b",
+        "B": "\\",
+        "e": "\033",
+        "f": "\f",
+        "n": "\n",
+        "r": "\r",
+        "t": "\t",
+        "v": "\v",
+        "0": "\0",
     }
 
+    _current_location = Location.nowhere
+    """
+    Location associated to the currently parsed regular expression (used to
+    emit diagnostics). Set by the ``nfa_for`` method.
+    """
+
     def __init__(self, case_insensitive: bool = False) -> None:
         """
         :param case_insensitive: Whether to consider regexps as case
@@ -249,8 +277,8 @@ class RegexpCollection:
             "[a-zA-Z]", etc.
         """
         self.case_insensitive = case_insensitive
-        self.patterns: Dict[str, RegexpCollection.Parser] = {}
-        self._visiting_patterns: Set[str] = set()
+        self.patterns: dict[str, RegexpCollection.Parser] = {}
+        self._visiting_patterns: set[str] = set()
 
     def _parse(self, regexp: str) -> RegexpCollection.Parser:
         """
@@ -262,23 +290,31 @@ class RegexpCollection:
 
         return root
 
-    def add_pattern(self, name: str, regexp: str) -> None:
+    def add_pattern(self, location: Location, name: str, regexp: str) -> None:
         """
         Add a named pattern to this collection.
 
+        :param location: Source location where the regular expression comes
+            from. Used to create diagnostics.
         :param name: Name for the pattern.
         :param regexp: Regular expression string for the pattern.
         """
         # Register the parser for regexp
         assert name not in self.patterns
+        RegexpCollection._current_location = location
         self.patterns[name] = self._parse(regexp)
 
-    def nfa_for(self, regexp: str) -> Tuple[NFAState, NFAState]:
+    def nfa_for(
+        self, location: Location, regexp: str
+    ) -> tuple[NFAState, NFAState]:
         """
         Parse the given regular expression string and return a NFA for it.
 
+        :param location: Source location where the regular expression comes
+            from. Used to create diagnostics.
         :param regexp: Regular expression to parse.
         """
+        RegexpCollection._current_location = location
         return self._parse(regexp).to_nfa(self)
 
     @contextmanager
@@ -295,7 +331,9 @@ class RegexpCollection:
         """
         check_source_language(
             rule_name not in self._visiting_patterns,
-            'infinite recursion in {}'.format(rule_name))
+            "infinite recursion in {}".format(rule_name),
+            location=self._current_location,
+        )
         self._visiting_patterns.add(rule_name)
         yield
         self._visiting_patterns.remove(rule_name)
@@ -320,25 +358,33 @@ class RegexpCollection:
 
         :param stream: Input regexp stream.
         """
-        assert stream.read() == '\\'
-        check_source_language(not stream.eof, 'bogus escape')
+        assert stream.read() == "\\"
+        check_source_language(
+            not stream.eof, "bogus escape", location=self._current_location
+        )
         char = stream.read()
 
         # If this encodes a Unicode code point, read the characters and turn
         # them into a Unicode character.
-        codepoint_chars_count = {'u': 4, 'U': 8}.get(char, None)
+        codepoint_chars_count = {"u": 4, "U": 8}.get(char, None)
         if codepoint_chars_count is not None:
             codepoint = 0
             for i in range(codepoint_chars_count):
-                check_source_language(not stream.eof, 'bogus Unicode escape')
+                check_source_language(
+                    not stream.eof,
+                    "bogus Unicode escape",
+                    location=self._current_location,
+                )
                 char = stream.read().lower()
-                if '0' <= char <= '9':
-                    digit = ord(char) - ord('0')
-                elif 'a' <= char <= 'f':
-                    digit = ord(char) - ord('a') + 0xa
+                if "0" <= char <= "9":
+                    digit = ord(char) - ord("0")
+                elif "a" <= char <= "f":
+                    digit = ord(char) - ord("a") + 0xA
                 else:
-                    check_source_language(
-                        False, 'invalid Unicode escape sequence')
+                    error(
+                        "invalid Unicode escape sequence",
+                        location=self._current_location,
+                    )
                 codepoint = codepoint * 16 + digit
             return codepoint
 
@@ -353,9 +399,9 @@ class RegexpCollection:
         codepoint = self._read_escape(stream)
         return self._char_set_for(codepoint)
 
-    def _parse_or(self,
-                  stream: SequenceReader,
-                  toplevel: bool = False) -> RegexpCollection.Parser:
+    def _parse_or(
+        self, stream: SequenceReader, toplevel: bool = False
+    ) -> RegexpCollection.Parser:
         """
         Read a sequence of alternatives. Stop at EOF or at the first unmatched
         parenthesis.
@@ -369,15 +415,20 @@ class RegexpCollection:
             subparsers.append(self._parse_sequence(stream))
             if stream.eof:
                 break
-            elif stream.next_is(')'):
-                check_source_language(not toplevel, 'unbalanced parentheses')
+            elif stream.next_is(")"):
+                check_source_language(
+                    not toplevel,
+                    "unbalanced parentheses",
+                    location=self._current_location,
+                )
                 break
             else:
-                assert stream.read() == '|'
+                assert stream.read() == "|"
         return self.Or(subparsers)
 
-    def _parse_sequence(self,
-                        stream: SequenceReader) -> RegexpCollection.Parser:
+    def _parse_sequence(
+        self, stream: SequenceReader
+    ) -> RegexpCollection.Parser:
         """
         Parse a sequence of regexps. Stop at the first unmatched parenthesis or
         at the first top-level pipe character.
@@ -386,42 +437,58 @@ class RegexpCollection:
         """
         subparsers = []
         while True:
-            if stream.eof or stream.next_is('|', ')'):
+            if stream.eof or stream.next_is("|", ")"):
                 break
 
-            elif stream.next_is('('):
+            elif stream.next_is("("):
                 # Nested group: recursively parse alternatives
                 stream.read()
                 subparsers.append(self._parse_or(stream))
-                check_source_language(stream.next_is(')'),
-                                      'unbalanced parenthesis')
+                check_source_language(
+                    stream.next_is(")"),
+                    "unbalanced parenthesis",
+                    location=self._current_location,
+                )
                 stream.read()
 
-            elif stream.next_is('['):
+            elif stream.next_is("["):
                 # Parse a range of characters
                 subparsers.append(self._parse_range(stream))
 
-            elif stream.next_is('{'):
+            elif stream.next_is("{"):
                 # Parse a reference to a named pattern (if the next character
                 # is a letter) or repeat the previous subparser otherwise.
                 stream.read()
-                name = ''
-                while not stream.eof and not stream.next_is('}'):
+                name = ""
+                while not stream.eof and not stream.next_is("}"):
                     name += stream.read()
-                check_source_language(stream.next_is('}'),
-                                      'unbalanced bracket')
+                check_source_language(
+                    stream.next_is("}"),
+                    "unbalanced bracket",
+                    location=self._current_location,
+                )
                 stream.read()
 
-                check_source_language(bool(name), "invalid empty brackets")
+                check_source_language(
+                    bool(name),
+                    "invalid empty brackets",
+                    location=self._current_location,
+                )
                 if name[0].isalpha():
-                    check_source_language(rule_name_re.match(name) is not None,
-                                          'invalid rule name: {}'.format(name))
+                    check_source_language(
+                        rule_name_re.match(name) is not None,
+                        "invalid rule name: {}".format(name),
+                        location=self._current_location,
+                    )
                     subparsers.append(self.Defer(name))
 
                 else:
                     m = repeat_re.match(name)
                     if m is None:
-                        error("invalid repetition specification")
+                        error(
+                            "invalid repetition specification",
+                            location=self._current_location,
+                        )
 
                     # "low" is mandatory, then even if there is a comma, high
                     # is optional (comma = repeat at least "low" times, no
@@ -437,6 +504,7 @@ class RegexpCollection:
                     check_source_language(
                         high is None or low <= high,
                         "min repeat greater than max repeat",
+                        location=self._current_location,
                     )
 
                     # Subparser to repeat
@@ -465,51 +533,63 @@ class RegexpCollection:
 
                     return self.Sequence(sequence)
 
-            elif stream.next_is('*', '+', '?'):
+            elif stream.next_is("*", "+", "?"):
                 # Repeat the previous sequence item
-                check_source_language(bool(subparsers), 'nothing to repeat')
+                check_source_language(
+                    bool(subparsers),
+                    "nothing to repeat",
+                    location=self._current_location,
+                )
                 check_source_language(
                     not isinstance(subparsers[-1], self.Repeat),
-                    'multiple repeat')
+                    "multiple repeat",
+                    location=self._current_location,
+                )
                 wrapper = {
-                    '*': lambda p: self.Repeat(p),
-                    '+': lambda p: self.Sequence([p, self.Repeat(p)]),
-                    '?': lambda p: self.Opt(p)
+                    "*": lambda p: self.Repeat(p),
+                    "+": lambda p: self.Sequence([p, self.Repeat(p)]),
+                    "?": lambda p: self.Opt(p),
                 }[stream.read()]
                 subparsers[-1] = wrapper(subparsers[-1])
 
-            elif stream.next_is('.'):
+            elif stream.next_is("."):
                 # Generally, "." designates any character *except* newlines. Do
                 # the same here.
                 stream.read()
-                subparsers.append(self.Range(CharSet('\n').negation))
+                subparsers.append(self.Range(CharSet("\n").negation))
 
-            elif stream.next_is('^', '$'):
-                check_source_language(
-                    False, 'matching beginning or ending is unsupported')
+            elif stream.next_is("^", "$"):
+                error(
+                    "matching beginning or ending is unsupported",
+                    location=self._current_location,
+                )
 
-            elif stream.next_is('\\'):
+            elif stream.next_is("\\"):
                 # Parse an escape sequence. In can be a Unicode character, a
                 # Unicode property or a simple escape sequence.
                 stream.read()
 
                 # \p and \P refer to character sets from Unicode general
                 # categories.
-                if stream.next_is('p', 'P'):
+                if stream.next_is("p", "P"):
                     action = stream.read()
 
                     # Read the category name, which must appear between curly
                     # brackets.
-                    category = ''
+                    category = ""
                     check_source_language(
-                        stream.next_is('{'),
-                        'incomplete Unicode category matcher')
+                        stream.next_is("{"),
+                        "incomplete Unicode category matcher",
+                        location=self._current_location,
+                    )
                     stream.read()
-                    while not stream.eof and not stream.next_is('}'):
+                    while not stream.eof and not stream.next_is("}"):
                         category += stream.read()
                     check_source_language(
-                        stream.next_is('}'),
-                        'incomplete Unicode category matcher')
+                        stream.next_is("}"),
+                        "incomplete Unicode category matcher",
+                        location=self._current_location,
+                    )
                     stream.read()
 
                     # If case insensitivity is enabled, the presence of either
@@ -529,20 +609,18 @@ class RegexpCollection:
                         try:
                             char_set = CharSet.for_category(category)
                         except KeyError:
-                            check_source_language(
-                                False,
-                                f'invalid Unicode category: {category}'
+                            error(
+                                f"invalid Unicode category: {category}",
+                                location=self._current_location,
                             )
 
-                    if action == 'P':
+                    if action == "P":
                         char_set = char_set.negation
                     subparsers.append(self.Range(char_set))
 
                 else:
                     stream.go_back()
-                    subparsers.append(
-                        self.Range(self._parse_escape(stream))
-                    )
+                    subparsers.append(self.Range(self._parse_escape(stream)))
 
             else:
                 char_set = self._char_set_for(ord(stream.read()))
@@ -556,12 +634,12 @@ class RegexpCollection:
 
         :param file stream: Input regexp stream.
         """
-        assert stream.read() == '['
-        ranges: List[Tuple[int, int]] = []
+        assert stream.read() == "["
+        ranges: list[tuple[int, int]] = []
 
         # First, determine if this range must be negated
         negate = False
-        if stream.next_is('^'):
+        if stream.next_is("^"):
             negate = True
             stream.read()
 
@@ -569,16 +647,21 @@ class RegexpCollection:
         #
         # TODO: handle '-' and ']' in first position.
         in_range = False
-        while not stream.eof and not stream.next_is(']'):
-            if stream.next_is('-'):
-                check_source_language(bool(ranges and not in_range),
-                                      'dangling dash')
+        while not stream.eof and not stream.next_is("]"):
+            if stream.next_is("-"):
+                check_source_language(
+                    bool(ranges and not in_range),
+                    "dangling dash",
+                    location=self._current_location,
+                )
                 in_range = True
                 stream.read()
             else:
-                codepoint = (self._read_escape(stream)
-                             if stream.next_is('\\')
-                             else ord(stream.read()))
+                codepoint = (
+                    self._read_escape(stream)
+                    if stream.next_is("\\")
+                    else ord(stream.read())
+                )
                 if in_range:
                     low, high = ranges.pop()
                     assert low == high
@@ -587,10 +670,17 @@ class RegexpCollection:
                     ranges.append((codepoint, codepoint))
                 in_range = False
 
-        check_source_language(not in_range, 'dangling dash')
-        check_source_language(stream.next_is(']'),
-                              'unbalanced square bracket')
-        assert stream.read() == ']'
+        check_source_language(
+            not in_range,
+            "dangling dash",
+            location=self._current_location,
+        )
+        check_source_language(
+            stream.next_is("]"),
+            "unbalanced square bracket",
+            location=self._current_location,
+        )
+        assert stream.read() == "]"
 
         # In case insensitivity is enabled, make sure both lowercase and
         # uppercase variants of all characters in ranges are present.
@@ -626,7 +716,7 @@ class NFAState:
         the corresponding DFA states.
         """
 
-        self.transitions: List[Tuple[Optional[CharSet], NFAState]] = []
+        self.transitions: list[tuple[CharSet | None, NFAState]] = []
         """
         List of associations between character sets and other states.
 
@@ -638,9 +728,9 @@ class NFAState:
         assert isinstance(other, NFAState)
         return self._id < other._id
 
-    def add_transition(self,
-                       chars: Optional[CharSet],
-                       next_state: NFAState) -> None:
+    def add_transition(
+        self, chars: CharSet | None, next_state: NFAState
+    ) -> None:
         """
         Add a transition from this state to another one.
 
@@ -656,15 +746,15 @@ class NFAState:
 
     @staticmethod
     def follow_spontaneous_transitions(
-        states: Iterable[NFAState]
-    ) -> Set[NFAState]:
+        states: Iterable[NFAState],
+    ) -> set[NFAState]:
         """
         Return the set of states that can be reached from the given set of
         states following spontaneous transitions.
 
         :param states: List of starting states.
         """
-        result: Set[NFAState] = set()
+        result: set[NFAState] = set()
 
         def process(state: NFAState) -> None:
             if state in result:
@@ -680,8 +770,8 @@ class NFAState:
 
     @staticmethod
     def reachable_nonspontaneous_transitions(
-        states: Tuple[NFAState, ...]
-    ) -> List[Tuple[CharSet, NFAState]]:
+        states: tuple[NFAState, ...]
+    ) -> list[tuple[CharSet, NFAState]]:
         """
         Return the list of non-spontaneous transitions that can be reached from
         the given set of states.
@@ -692,7 +782,7 @@ class NFAState:
 
         :param states: List of starting states.
         """
-        result: List[Tuple[CharSet, NFAState]] = []
+        result: list[tuple[CharSet, NFAState]] = []
         for state in states:
             for chars, next_state in state.transitions:
                 if chars is None:
@@ -702,7 +792,7 @@ class NFAState:
         return result
 
     @staticmethod
-    def hashable_state_set(states: Set[NFAState]) -> Tuple[NFAState, ...]:
+    def hashable_state_set(states: set[NFAState]) -> tuple[NFAState, ...]:
         """
         Return a value that can be used as a dict key/set element to represent
         a set of states.
@@ -711,8 +801,8 @@ class NFAState:
 
     @staticmethod
     def deterministic_transitions(
-        states: Tuple[NFAState, ...]
-    ) -> Dict[Tuple[NFAState, ...], CharSet]:
+        states: tuple[NFAState, ...]
+    ) -> dict[tuple[NFAState, ...], CharSet]:
         """
         Return the set of deterministic (non-spontaneous and disjoint)
         transitions that leave the "states" sub-graph.
@@ -728,7 +818,7 @@ class NFAState:
         """
         # First, collect for each reachable state the set of characters that
         # allow to reach that state.
-        transitions: Dict[NFAState, CharSet] = {}
+        transitions: dict[NFAState, CharSet] = {}
         for chars, next_state in NFAState.reachable_nonspontaneous_transitions(
             states
         ):
@@ -759,10 +849,10 @@ class NFAState:
         # }.
         class Event:
             def __init__(self) -> None:
-                self.adding: Set[NFAState] = set()
-                self.removing: Set[NFAState] = set()
+                self.adding: set[NFAState] = set()
+                self.removing: set[NFAState] = set()
 
-        event_map: Dict[int, Event] = defaultdict(Event)
+        event_map: dict[int, Event] = defaultdict(Event)
         for next_state, chars in transitions.items():
             for r in chars.ranges:
                 event_map[r[0]].adding.add(next_state)
@@ -772,11 +862,11 @@ class NFAState:
         # The final step is to compute the set of transitions for which
         # character sets are disjoint: just follow the stream of events.
 
-        result: Dict[Tuple[NFAState, ...], CharSet] = defaultdict(CharSet)
+        result: dict[tuple[NFAState, ...], CharSet] = defaultdict(CharSet)
 
-        def add_transition(low: Optional[int],
-                           high: int,
-                           states: Iterable[NFAState]) -> None:
+        def add_transition(
+            low: int | None, high: int, states: Iterable[NFAState]
+        ) -> None:
             if states:
                 assert isinstance(low, int)
             else:
@@ -787,10 +877,10 @@ class NFAState:
             )
 
         # Set of states "active" for the current position in the events stream
-        active_states: Set[NFAState] = set()
+        active_states: set[NFAState] = set()
 
         # Character for the last event we processed
-        last_char: Optional[int] = None
+        last_char: int | None = None
 
         for char, event in events:
             assert event.adding or event.removing
@@ -811,17 +901,17 @@ class NFAState:
         """
         Return the conversion of this NFA into a DFA.
         """
-        result: Optional[DFAState] = None
+        result: DFAState | None = None
 
         # Mapping from sets of NFAState nodes (see NFAState.hashable_state_set)
         # to the corresponding DFAState nodes.
-        dfa_states: Dict[Tuple[NFAState, ...], DFAState] = {}
+        dfa_states: dict[tuple[NFAState, ...], DFAState] = {}
 
         # List of tuple[NFAState, CharSet and tuple[NFAState]] for the
         # transitions that constitute the DFA.
-        transitions: List[Tuple[Tuple[NFAState, ...],
-                                CharSet,
-                                Tuple[NFAState, ...]]] = []
+        transitions: list[
+            tuple[tuple[NFAState, ...], CharSet, tuple[NFAState, ...]]
+        ] = []
 
         queue = {
             self.hashable_state_set(
@@ -834,8 +924,9 @@ class NFAState:
                 # TODO: unreachable?
                 raise RuntimeError()
 
-            new_node = DFAState(labels={s.label for s in states
-                                        if s.label is not None})
+            new_node = DFAState(
+                labels={s.label for s in states if s.label is not None}
+            )
             dfa_states[states] = new_node
             if result is None:
                 result = new_node
@@ -848,8 +939,9 @@ class NFAState:
                 transitions.append((states, char_set, next_states))
 
         for states, char_set, next_states in transitions:
-            dfa_states[states].add_transition(char_set,
-                                              dfa_states[next_states])
+            dfa_states[states].add_transition(
+                char_set, dfa_states[next_states]
+            )
 
         assert result
         return result
@@ -866,14 +958,14 @@ class DFAState:
     Single state in a deterministic state machine.
     """
 
-    def __init__(self, labels: Optional[Set[Any]] = None):
+    def __init__(self, labels: set[Any] | None = None):
         self.labels = labels or set()
         """
         Set of labels inheritted from the set of NFA states that this DFA state
         represents.
         """
 
-        self.transitions: List[Tuple[CharSet, DFAState]] = []
+        self.transitions: list[tuple[CharSet, DFAState]] = []
         """
         List of associations between character sets and other states.
 
@@ -895,9 +987,10 @@ class DFAState:
         # Check that ``chars`` does overlap with character sets for other
         # transitions.
         for other_chars, _ in self.transitions:
-            assert not chars.overlaps_with(other_chars), (
-                'Overlapping input char sets: {} and {}'
-                .format(chars, other_chars)
+            assert not chars.overlaps_with(
+                other_chars
+            ), "Overlapping input char sets: {} and {}".format(
+                chars, other_chars
             )
 
         self.transitions.append((chars, next_state))
@@ -906,9 +999,11 @@ class DFAState:
         """
         Return a dot script representing this DFA.
         """
-        return _to_dot(self,
-                       lambda s: s.transitions,
-                       lambda s: '\n'.join(str(l) for l in sorted(s.labels)))
+        return _to_dot(
+            self,
+            lambda s: s.transitions,
+            lambda s: "\n".join(str(l) for l in sorted(s.labels)),
+        )
 
 
 class DFACodeGenHolder:
@@ -917,11 +1012,13 @@ class DFACodeGenHolder:
     """
 
     class State:
-        def __init__(self,
-                     dfa_state: DFAState,
-                     label: str,
-                     transitions: List[Tuple[CharSet, DFAState]],
-                     action: Optional[RuleAssoc]):
+        def __init__(
+            self,
+            dfa_state: DFAState,
+            label: str,
+            transitions: list[tuple[CharSet, DFAState]],
+            action: RuleAssoc | None,
+        ):
             self.dfa_state = dfa_state
             """
             DFA state this represents.
@@ -955,13 +1052,13 @@ class DFACodeGenHolder:
 
             self._transitions = transitions
 
-            self.case_transitions: List[Tuple[CharSet, str]] = []
+            self.case_transitions: list[tuple[CharSet, str]] = []
             """
             List of transitions to be lowered to a case statement. This maps
             maching characters to code labels for the next states.
             """
 
-            self.table_transitions: List[Tuple[CharSet, str]] = []
+            self.table_transitions: list[tuple[CharSet, str]] = []
             """
             List of transitions to be lowered to table-based character lookups.
             This maps the character sets for the lookup table for character
@@ -969,15 +1066,16 @@ class DFACodeGenHolder:
             characters.
             """
 
-            self.named_table_transitions: List[Tuple[str, str]] = []
+            self.named_table_transitions: list[tuple[str, str]] = []
             """
             List of table-based transitions. This contains the same data as
             ``table_transitions``, but with table names (for generated code
             entities) rather than abstract character sets.
             """
 
-        def compute_transitions(self,
-                                state_labels: Dict[DFAState, str]) -> None:
+        def compute_transitions(
+            self, state_labels: dict[DFAState, str]
+        ) -> None:
             """
             Compute self.case_transitions and self.table_transitions.
 
@@ -1001,10 +1099,10 @@ class DFACodeGenHolder:
         def has_transitions(self) -> bool:
             return bool(self.case_transitions or self.table_transitions)
 
-    def __init__(self,
-                 dfa: DFAState,
-                 get_action: Callable[[Set[Any]], Optional[RuleAssoc]]):
-        self.states: List[DFACodeGenHolder.State] = []
+    def __init__(
+        self, dfa: DFAState, get_action: Callable[[set[Any]], RuleAssoc | None]
+    ):
+        self.states: list[DFACodeGenHolder.State] = []
 
         # Compute the list of states corresponding to the code blocks to emit.
         # We store them in a list (self.states) to have deterministic code
@@ -1028,9 +1126,9 @@ class DFACodeGenHolder:
             self.states.append(
                 self.State(
                     dfa_state,
-                    f'State_{len(self.states)}',
+                    f"State_{len(self.states)}",
                     transitions,
-                    get_action(dfa_state.labels)
+                    get_action(dfa_state.labels),
                 )
             )
 
@@ -1045,15 +1143,16 @@ class DFACodeGenHolder:
         # Generate arrays for table-based character lookups.  At the same time,
         # replace character sets with these names in table-based transitions.
 
-        self.charset_to_tablename: Dict[CharSet, str] = {}
+        self.charset_to_tablename: dict[CharSet, str] = {}
 
         for state in self.states:
             for char_set, label in state.table_transitions:
                 try:
                     table_name = self.charset_to_tablename[char_set]
                 except KeyError:
-                    table_name = 'Ranges_{}'.format(
-                        len(self.charset_to_tablename))
+                    table_name = "Ranges_{}".format(
+                        len(self.charset_to_tablename)
+                    )
                     self.charset_to_tablename[char_set] = table_name
                 state.named_table_transitions.append((table_name, label))
 
@@ -1066,16 +1165,19 @@ class DFACodeGenHolder:
             for char_set, table_name in self.charset_to_tablename.items()
         )
 
-        lines: List[str] = []
+        lines: list[str] = []
         for table_name, char_set in tables:
-            ranges: List[str] = []
+            ranges: list[str] = []
             for l, h in char_set.ranges:
                 if ranges:
-                    ranges[-1] += ','
-                ranges.append("(Character_Type'Val ({}),"
-                              " Character_Type'Val ({}))".format(l, h))
-            lines.append('{} : constant Character_Range_Array := ('
-                         .format(table_name))
+                    ranges[-1] += ","
+                ranges.append(
+                    "(Character_Type'Val ({}),"
+                    " Character_Type'Val ({}))".format(l, h)
+                )
+            lines.append(
+                "{} : constant Character_Range_Array := (".format(table_name)
+            )
             lines.extend(ranges)
-            lines.append(');')
-        return '\n'.join(prefix + line for line in lines)
+            lines.append(");")
+        return "\n".join(prefix + line for line in lines)
